searchNodes=[{"ref":"Oban.html","title":"Oban","type":"module","doc":"Oban isn&#39;t an application and won&#39;t be started automatically. It is started by a supervisor that must be included in your application&#39;s supervision tree. All of your configuration is passed into the Oban supervisor, allowing you to configure Oban like the rest of your application. # confg/config.exs config :my_app, Oban, repo: MyApp.Repo, queues: [default: 10, events: 50, media: 20] # lib/my_app/application.ex defmodule MyApp.Application do @moduledoc false use Application alias MyApp.{Endpoint, Repo} def start(_type, _args) do children = [ Repo, Endpoint, {Oban, Application.get_env(:my_app, Oban)} ] Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor) end end Configuring Queues Queues are specified as a keyword list where the key is the name of the queue and the value is the maximum number of concurrent jobs. The following configuration would start four queues with concurrency ranging from 5 to 50: queues: [default: 10, mailers: 20, events: 50, media: 5] There isn&#39;t a limit to the number of queues or how many jobs may execute concurrently. Here are a few caveats and guidelines: Each queue will run as many jobs as possible concurrently, up to the configured limit. Make sure your system has enough resources (i.e. database connections) to handle the concurrent load. Only jobs in the configured queues will execute. Jobs in any other queue will stay in the database untouched. Be careful how many concurrent jobs make expensive system calls (i.e. FFMpeg, ImageMagick). The BEAM ensures that the system stays responsive under load, but those guarantees don&#39;t apply when using ports or shelling out commands. Defining Workers Worker modules do the work of processing a job. At a minimum they must define a perform/2 function, which is called first with the full Oban.Job struct, and subsequently with the args map if no clause matches. Define a worker to process jobs in the events queue: defmodule MyApp.Business do use Oban.Worker, queue: &quot;events&quot;, max_attempts: 10 @impl Worker def perform(%{&quot;id&quot; =&gt; id}, _job) do model = MyApp.Repo.get(MyApp.Business.Man, id) IO.inspect(model) end end The value returned from perform/2 is ignored, unless it an {:error, reason} tuple. With an error return or when perform has an uncaught exception or throw then the error will be reported and the job will be retried (provided there are attempts remaining). See Oban.Worker for more details on failure conditions and Oban.Telemetry for details on job reporting. Enqueueing Jobs Jobs are simply Ecto structs and are enqueued by inserting them into the database. For convenience and consistency all workers provide a new/2 function that converts an args map into a job changeset suitable for insertion: %{in_the: &quot;business&quot;, of_doing: &quot;business&quot;} |&gt; MyApp.Business.new() |&gt; Oban.insert() The worker&#39;s defaults may be overridden by passing options: %{vote_for: &quot;none of the above&quot;} |&gt; MyApp.Business.new(queue: &quot;special&quot;, max_attempts: 5) |&gt; Oban.insert() Jobs may be scheduled at a specific datetime in the future: %{id: 1} |&gt; MyApp.Business.new(scheduled_at: ~U[2020-12-25 19:00:56.0Z]) |&gt; Oban.insert() Jobs may also be scheduled down to the second any time in the future: %{id: 1} |&gt; MyApp.Business.new(schedule_in: 5) |&gt; Oban.insert() Unique jobs can be configured in the worker, or when the job is built: %{email: &quot;brewster@example.com&quot;} |&gt; MyApp.Mailer.new(unique: [period: 300, fields: [:queue, :worker]) |&gt; Oban.insert() Multiple jobs can be inserted in a single transaction: Ecto.Multi.new() |&gt; Oban.insert(:b_job, MyApp.Business.new(%{id: 1})) |&gt; Oban.insert(:m_job, MyApp.Mailer.new(%{email: &quot;brewser@example.com&quot;})) |&gt; Repo.transaction() Occasionally you may need to insert a job for a worker that exists in another application. In that case you can use Oban.Job.new/2 to build the changeset manually: %{id: 1, user_id: 2} |&gt; Oban.Job.new(queue: :default, worker: OtherApp.Worker) |&gt; Oban.insert() Oban.insert/2,4 is the preferred way of inserting jobs as it provides some of Oban&#39;s advanced features (i.e., unique jobs). However, you can use your application&#39;s Repo.insert/2 function if necessary. See Oban.Job.new/2 for a full list of job options. Unique Jobs The unique jobs feature lets you specify constraints to prevent enqueuing duplicate jobs. Uniquness is based on a combination of args, queue, worker, state and insertion time. It is configured at the worker or job level using the following options: :period — The number of seconds until a job is no longer considered duplicate. You should always specify a period. :fields — The fields to compare when evaluating uniqueness. The available fields are :args, :queue and :worker, by default all three are used. :states — The job states that will be checked for duplicates. The available states are :available, :scheduled, :executing, :retryable and :completed. By default all states are checked, which prevents any duplicates, even if the previous job has been completed. For example, configure a worker to be unique across all fields and states for 60 seconds: use Oban.Worker, unique: [period: 60] Configure the worker to be unique only by :worker and :queue: use Oban.Worker, unique: [fields: [:queue, :worker], period: 60] Or, configure a worker to be unique until it has executed: use Oban.Worker, unique: [period: 300, states: [:available, :scheduled, :executing]] Stronger Guarantees Oban&#39;s unique job support is built on a client side read/write cycle. That makes it subject to duplicate writes if two transactions are started simultaneously. If you absolutely must ensure that a duplicate job isn&#39;t inserted then you will have to make use of unique constraints within the database. Oban.insert/2,4 will handle unique constraints safely through upsert support. Performance Note If your application makes heavy use of unique jobs you may want to add indexes on the args and inserted_at columns of the oban_jobs table. The other columns considered for uniqueness are already covered by indexes. Testing As noted in the Usage section above there are some guidelines for running tests: Disable all job dispatching by setting queues: false or queues: nil in your test.exs config. Keyword configuration is deep merged, so setting queues: [] won&#39;t have any effect. Disable pruning via prune: :disabled. Pruning isn&#39;t necessary in testing mode because jobs created within the sandbox are rolled back at the end of the test. Additionally, the periodic pruning queries will raise DBConnection.OwnershipError when the application boots. Be sure to use the Ecto Sandbox for testing. Oban makes use of database pubsub events to dispatch jobs, but pubsub events never fire within a transaction. Since sandbox tests run within a transaction no events will fire and jobs won&#39;t be dispatched. config :my_app, MyApp.Repo, pool: Ecto.Adapters.SQL.Sandbox Oban provides some helpers to facilitate testing. The helpers handle the boilerplate of making assertions on which jobs are enqueued. To use the assert_enqueued/1 and refute_enqueued/1 helpers in your tests you must include them in your testing module and specify your app&#39;s Ecto repo: use Oban.Testing, repo: MyApp.Repo Now you can assert, refute or list jobs that have been enqueued within your tests: assert_enqueued worker: MyWorker, args: %{id: 1} # or refute_enqueued queue: &quot;special&quot;, args: %{id: 2} # or assert [%{args: %{&quot;id&quot; =&gt; 1}}] = all_enqueued worker: MyWorker See the Oban.Testing module for more details on making assertions. Integration Testing During integration testing it may be necessary to run jobs because they do work essential for the test to complete, i.e. sending an email, processing media, etc. You can execute all available jobs in a particular queue by calling Oban.drain_queue/1 directly from your tests. For example, to process all pending jobs in the &quot;mailer&quot; queue while testing some business logic: defmodule MyApp.BusinessTest do use MyApp.DataCase, async: true alias MyApp.{Business, Worker} test &quot;we stay in the business of doing business&quot; do :ok = Business.schedule_a_meeting(%{email: &quot;monty@brewster.com&quot;}) assert %{success: 1, failure: 0} == Oban.drain_queue(:mailer) # Make an assertion about the email delivery end end See Oban.drain_queue/1 for additional details. Error Handling When a job returns an error value, raises an error or exits during execution the details are recorded within the errors array on the job. When the number of execution attempts is below the configured max_attempts limit, the job will automatically be retried in the future. The retry delay has an exponential backoff, meaning the job&#39;s second attempt will be after 16s, third after 31s, fourth after 1m 36s, etc. See the Oban.Worker documentation on &quot;Customizing Backoff&quot; for alternative backoff strategies. Error Details Execution errors are stored as a formatted exception along with metadata about when the failure ocurred and which attempt caused it. Each error is stored with the following keys: at The utc timestamp when the error occurred at attempt The attempt number when the error ocurred error A formatted error message and stacktrace See the Instrumentation docs below for an example of integrating with external error reporting systems. Limiting Retries By default jobs will be retried up to 20 times. The number of retries is controlled by the max_attempts value, which can be set at the Worker or Job level. For example, to instruct a worker to discard jobs after three failures: use Oban.Worker, queue: &quot;limited&quot;, max_attempts: 3 Pruning Historic Jobs Job stats and queue introspection is built on keeping job rows in the database after they have completed. This allows administrators to review completed jobs and build informative aggregates, but at the expense of storage and an unbounded table size. To prevent the oban_jobs table from growing indefinitely, Oban provides active pruning of completed jobs. By default, pruning is disabled. To enable pruning we configure a supervision tree with the :prune option. There are three distinct modes of pruning: :disabled - This is the default, where no pruning happens at all {:maxlen, count} - Pruning is based on the number of rows in the table, any rows beyond the configured count will be deleted {:maxage, seconds} - Pruning is based on a row&#39;s age, any rows older than the configured number of seconds will be deleted. The age unit is always specified in seconds, but values on the scale of days, weeks or months are perfectly acceptable. Pruning is best-effort and performed out-of-band. This means that all limits are soft; jobs beyond a specified length or age may not be pruned immediately after jobs complete. Prune timing is based on the configured prune_interval, which is one minute by default. Only jobs in a completed or discarded state will be deleted. Currently executing jobs and older jobs that are still in the available state will be retained. Instrumentation &amp; Logging Oban provides integration with Telemetry, a dispatching library for metrics. It is easy to report Oban metrics to any backend by attaching to :oban events. For instructions on using the default structured logger and information on event metadata see docs for the Oban.Telemetry module. Prefix Support Oban supports namespacing through PostgreSQL schemas, also called &quot;prefixes&quot; in Ecto. With prefixes your jobs table can reside outside of your primary schema (usually public) and you can have multiple separate job tables. To use a prefix you first have to specify it within your migration: defmodule MyApp.Repo.Migrations.AddPrefixedObanJobsTable do use Ecto.Migration def up do Oban.Migrations.up(prefix: &quot;private&quot;) end def down do Oban.Migrations.down(prefix: &quot;private&quot;) end end The migration will create the &quot;private&quot; schema and all tables, functions and triggers within that schema. With the database migrated you&#39;ll then specify the prefix in your configuration: config :my_app, Oban, prefix: &quot;private&quot;, repo: MyApp.Repo, queues: [default: 10] Now all jobs are inserted and executed using the private.oban_jobs table. Note that Oban.insert/2,4 will write jobs in the private.oban_jobs table, you&#39;ll need to specify a prefix manually if you insert jobs directly through a repo. Isolation Not only is the oban_jobs table isolated within the schema, but all notification events are also isolated. That means that insert/update events will only dispatch new jobs for their prefix. You can run multiple Oban instances with different prefixes on the same system and have them entirely isolated, provided you give each supervisor a distinct id. Here we configure our application to start three Oban supervisors using the &quot;public&quot;, &quot;special&quot; and &quot;private&quot; prefixes, respectively: def start(_type, _args) do children = [ Repo, Endpoint, Supervisor.child_spec({Oban, name: ObanA, repo: Repo}, id: ObanA), Supervisor.child_spec({Oban, name: ObanB, repo: Repo, prefix: &quot;special&quot;}, id: ObanB), Supervisor.child_spec({Oban, name: ObanC, repo: Repo, prefix: &quot;private&quot;}, id: ObanC) ] Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor) end"},{"ref":"Oban.html#child_spec/1","title":"Oban.child_spec/1","type":"function","doc":"Returns a specification to start this module under a supervisor. See Supervisor."},{"ref":"Oban.html#config/1","title":"Oban.config/1","type":"function","doc":"Retreive the config struct for a named Oban supervision tree."},{"ref":"Oban.html#drain_queue/2","title":"Oban.drain_queue/2","type":"function","doc":"Synchronously execute all available jobs in a queue. All execution happens within the current process and it is guaranteed not to raise an error or exit. Draining a queue from within the current process is especially useful for testing. Jobs that are enqueued by a process when Ecto is in sandbox mode are only visible to that process. Calling drain_queue/1 allows you to control when the jobs are executed and to wait synchronously for all jobs to complete. Failures &amp; Retries Draining a queue uses the same execution mechanism as regular job dispatch. That means that any job failures or crashes will be captured and result in a retry. Retries are scheduled in the future with backoff and won&#39;t be retried immediately. Exceptions are not raised in to the calling process. If you expect jobs to fail, would like to track failures, or need to check for specific errors you can use one of these mechanisms: Check for side effects from job execution Use telemetry events to track success and failure Check the database for jobs with errors Example Drain a queue with three available jobs, two of which succeed and one of which fails: Oban.drain_queue(:default) %{success: 2, failure: 1}"},{"ref":"Oban.html#insert/2","title":"Oban.insert/2","type":"function","doc":"Insert a new job into the database for execution. This and the other insert variants are the recommended way to enqueue jobs because they support features like unique jobs. See the section on &quot;Unique Jobs&quot; for more details. Example Insert a single job: {:ok, job} = Oban.insert(MyApp.Worker.new(%{id: 1})) Insert a job while ensuring that it is unique within the past 30 seconds: {:ok, job} = Oban.insert(MyApp.Worker.new(%{id: 1}, unique: [period: 30]))"},{"ref":"Oban.html#insert/4","title":"Oban.insert/4","type":"function","doc":"Put a job insert operation into an Ecto.Multi. Like insert/2, this variant is recommended over Ecto.Multi.insert beause it supports all of Oban&#39;s features, i.e. unique jobs. See the section on &quot;Unique Jobs&quot; for more details. Example Ecto.Multi.new() |&gt; Oban.insert(:job_1, MyApp.Worker.new(%{id: 1})) |&gt; Oban.insert(:job_2, MyApp.Worker.new(%{id: 2})) |&gt; MyApp.Repo.transaction()"},{"ref":"Oban.html#insert!/2","title":"Oban.insert!/2","type":"function","doc":"Similar to insert/2, but raises an Ecto.InvalidChangesetError if the job can&#39;t be inserted. Example job = Oban.insert!(MyApp.Worker.new(%{id: 1}))"},{"ref":"Oban.html#insert_all/2","title":"Oban.insert_all/2","type":"function","doc":"Insert multiple jobs into the database for execution. Insertion respects prefix and verbose settings, but it does not use per-job unique configuration. You must use insert/2,4 or insert!/2 for per-job unique support. There are a few important differences between this function and Ecto.Repo.insert_all/3: This function always returns a list rather than a tuple of {count, records} This function requires a list of changesets rather than a list of maps or keyword lists Example 1..100 |&gt; Enum.map(&amp;MyApp.Worker.new(%{id: &amp;1})) |&gt; Oban.insert_all()"},{"ref":"Oban.html#insert_all/4","title":"Oban.insert_all/4","type":"function","doc":"Put an insert_all operation into an Ecto.Multi. This function supports the same features and has the same caveats as insert_all/2. Example changesets = Enum.map(0..100, MyApp.Worker.new(%{id: &amp;1})) Ecto.Multi.new() |&gt; Oban.insert_all(:jobs, changesets) |&gt; MyApp.Repo.transaction()"},{"ref":"Oban.html#kill_job/2","title":"Oban.kill_job/2","type":"function","doc":"Kill an actively executing job and mark it as discarded, ensuring that it won&#39;t be retried. If the job happens to fail before it can be killed the state is set to discarded. However, if it manages to complete successfully then the state will still be completed. Example Kill a long running job with an id of 1: Oban.kill_job(1) :ok"},{"ref":"Oban.html#pause_queue/2","title":"Oban.pause_queue/2","type":"function","doc":"Pause a running queue, preventing it from executing any new jobs. All running jobs will remain running until they are finished. When shutdown begins all queues are paused. Example Pause the default queue: Oban.pause_queue(:default) :ok"},{"ref":"Oban.html#resume_queue/2","title":"Oban.resume_queue/2","type":"function","doc":"Resume executing jobs in a paused queue. Example Resume a paused default queue: Oban.resume_queue(:default) :ok"},{"ref":"Oban.html#scale_queue/3","title":"Oban.scale_queue/3","type":"function","doc":"Scale the concurrency for a queue. Example Scale a queue up, triggering immediate execution of queued jobs: Oban.scale_queue(:default, 50) :ok Scale the queue back down, allowing executing jobs to finish: Oban.scale_queue(:default, 5) :ok"},{"ref":"Oban.html#start_link/1","title":"Oban.start_link/1","type":"function","doc":"Starts an Oban supervision tree linked to the current process. Options :name — used for name supervisor registration :node — used to identify the node that the supervision tree is running in. If no value is provided it will use the node name in a distributed system, the hostname in an isolated node. See &quot;Node Name&quot; below. :repo — specifies the Ecto repo used to insert and retreive jobs. :queues — a keyword list where the keys are queue names and the values are the concurrency setting. For example, setting queues to [default: 10, exports: 5] would start the queues default and exports with a combined concurrency level of 20. The concurrency setting specifies how many jobs each queue will run concurrently. For testing purposes :queues may be set to false or nil, which effectively disables all job dispatching. :prefix — the query prefix, or schema, to use for inserting and executing jobs. An oban_jobs table must exist within the prefix. See the &quot;Prefix Support&quot; section in the module documentation for more details. :poll_interval - the number of milliseconds between polling for new jobs in a queue. This is directly tied to the resolution of scheduled jobs. For example, with a poll_interval of 5_000ms, scheduled jobs are checked every 5 seconds. The default is 1_000ms. :prune - configures job pruning behavior, see &quot;Pruning Historic Jobs&quot; for more information :prune_interval — the number of milliseconds between calls to prune historic jobs. The default is 60_000ms, or one minute. :prune_limit – the maximum number of jobs that will be pruned at each prune interval. The default is 5_000. :rescue_after — the number of seconds after an executing job without any pulse activity may be rescued. This value must be greater than the poll_interval. The default is 60s. :rescue_interval — the number of milliseconds between calls to rescue orphaned jobs, the default is 60_000ms, or one minute. :shutdown_grace_period - the amount of time a queue will wait for executing jobs to complete before hard shutdown, specified in milliseconds. The default is 15_000, or 15 seconds. :verbose — determines whether queries will be logged or not, does not supersede the repo&#39;s configured log level. Defaults to true, where all queries will be logged. Note that any options passed to start_link will override options set through the using macro. Examples To start an Oban supervisor within an application&#39;s supervision tree: def start(_type, _args) do children = [MyApp.Repo, {Oban, queues: [default: 50]}] Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor) end Node Name When the node value hasn&#39;t been configured it will be generated based on the environment: In a distributed system the node name is used In a Heroku environment the system environment&#39;s DYNO value is used Otherwise, the system hostname is used"},{"ref":"Oban.html#t:option/0","title":"Oban.option/0","type":"type","doc":""},{"ref":"Oban.Job.html","title":"Oban.Job","type":"module","doc":"A Job is an Ecto schema used for asynchronous execution. Job changesets are created by your application code and inserted into the database for asynchronous execution. Jobs can be inserted along with other application data as part of a transaction, which guarantees that jobs will only be triggered from a successful transaction."},{"ref":"Oban.Job.html#new/2","title":"Oban.Job.new/2","type":"function","doc":"Construct a new job changeset ready for insertion into the database. Options :max_attempts — the maximum number of times a job can be retried if there are errors during execution :queue — a named queue to push the job into. Jobs may be pushed into any queue, regardless of whether jobs are currently being processed for the queue. :schedule_in - the number of seconds until the job should be executed :scheduled_at - a time in the future after which the job should be executed :worker — a module to execute the job in. The module must implement the Oban.Worker behaviour. :unique — a keyword list of options specifying how uniquness will be calculated. The options define which fields will be used, for how long, and for which states. Examples Insert a job with the :default queue: %{id: 1, user_id: 2} |&gt; Oban.Job.new(queue: :default, worker: MyApp.Worker) |&gt; Oban.insert() Generate a pre-configured job for MyApp.Worker and push it: %{id: 1, user_id: 2} |&gt; MyApp.Worker.new() |&gt; Oban.insert() Schedule a job to run in 5 seconds: %{id: 1} |&gt; MyApp.Worker.new(schedule_in: 5) |&gt; Oban.insert() Insert a job, ensuring that it is unique within the past minute: %{id: 1} |&gt; MyApp.Worker.new(unique: [period: 60]) |&gt; Oban.insert() Insert a unique job based only on the worker field, and within multiple states: fields = [:worker] states = [:available, :scheduled, :executing, :retryable, :completed] %{id: 1} |&gt; MyApp.Worker.new(unique: [fields: fields, period: 60, states: states]) |&gt; Oban.insert()"},{"ref":"Oban.Job.html#to_map/1","title":"Oban.Job.to_map/1","type":"function","doc":"Convert a Job changeset into a map suitable for database insertion. Examples Convert a worker generated changeset into a plain map: %{id: 123} |&gt; MyApp.Worker.new() |&gt; Oban.Job.to_map()"},{"ref":"Oban.Job.html#t:args/0","title":"Oban.Job.args/0","type":"type","doc":""},{"ref":"Oban.Job.html#t:errors/0","title":"Oban.Job.errors/0","type":"type","doc":""},{"ref":"Oban.Job.html#t:option/0","title":"Oban.Job.option/0","type":"type","doc":""},{"ref":"Oban.Job.html#t:t/0","title":"Oban.Job.t/0","type":"type","doc":""},{"ref":"Oban.Job.html#t:unique_field/0","title":"Oban.Job.unique_field/0","type":"type","doc":""},{"ref":"Oban.Job.html#t:unique_option/0","title":"Oban.Job.unique_option/0","type":"type","doc":""},{"ref":"Oban.Job.html#t:unique_state/0","title":"Oban.Job.unique_state/0","type":"type","doc":""},{"ref":"Oban.Telemetry.html","title":"Oban.Telemetry","type":"module","doc":"Telemetry integration for event metrics, logging and error reporting. Oban currently emits an event when a job has exeucted: [:oban, :success] if the job succeeded or [:oban, :failure] if there was an error or the process crashed. All job events share the same details about the job that was executed. In addition, failed jobs provide the error type, the error itself, and the stacktrace. The following chart shows which metadata you can expect for each event: eventmetadata :success:id, :args, :queue, :worker, :attempt, :max_attempt :failure:id, :args, :queue, :worker, :attempt, :max_attempt, :kind, :error, :stack For :failure events the metadata will include details about what caused the failure. The :kind value is determined by how an error occurred. Here are the possible kinds: :error — from an {:error, error} return value. Some Erlang functions may also throw an :error tuple, which will be reported as :error. :exception — from a rescued exception :exit — from a caught process exit :throw — from a caught value, this doesn&#39;t necessarily mean that an error occurred and the error value is unpredictable Default Logger A default log handler that emits structured JSON is provided, see attach_default_logger/0 for usage. Otherwise, if you would prefer more control over logging or would like to instrument events you can write your own handler. Examples A handler that only logs a few details about failed jobs: defmodule MicroLogger do require Logger def handle_event([:oban, :failure], %{duration: duration}, meta, nil) do Logger.warning(&quot;[\#{meta.queue}] \#{meta.worker} failed in \#{duration}&quot;) end end :telemetry.attach(&quot;oban-logger&quot;, [:oban, :failure], &amp;ObanLogger.handle_event/4, nil) Another great use of execution data is error reporting. Here is an example of integrating with Honeybadger, but only reporting jobs that have failed 3 times or more: defmodule ErrorReporter do def handle_event([:oban, :failure], _timing, %{attempt: attempt} = meta, nil) do if attempt &gt;= 3 do context = Map.take(meta, [:id, :args, :queue, :worker]) Honeybadger.notify(meta.error, context, meta.stack) end end end :telemetry.attach(&quot;oban-errors&quot;, [:oban, :failure], &amp;ErrorReporter.handle_event/4, nil)"},{"ref":"Oban.Telemetry.html#attach_default_logger/1","title":"Oban.Telemetry.attach_default_logger/1","type":"function","doc":"Attaches a default structured JSON Telemetry handler for logging. This function attaches a handler that outputs logs with the following fields: source — always &quot;oban&quot; event — either :success or :failure dependening on whether the job succeeded or errored args — a map of the job&#39;s raw arguments worker — the job&#39;s worker module queue — the job&#39;s queue duration — the job&#39;s runtime duration in microseconds Examples Attach a logger at the default :info level: :ok = Oban.Telemetry.attach_default_logger() Attach a logger at the :debug level: :ok = Oban.Telemetry.attach_default_logger(:debug)"},{"ref":"Oban.Testing.html","title":"Oban.Testing","type":"module","doc":"This module simplifies making assertions about enqueued jobs during testing. Assertions may be made on any property of a job, but you&#39;ll typically want to check by args, queue or worker. Using in Tests The most convenient way to use Oban.Testing is to use the module: use Oban.Testing, repo: MyApp.Repo That will define three helper functions, assert_enqueued/1, refute_enqueued/1 and all_enqueued/1. The functions can then be used to make assertions on the jobs that have been inserted in the database while testing. assert_enqueued worker: MyWorker, args: %{id: 1} # or refute_enqueued queue: &quot;special&quot;, args: %{id: 2} # or assert [%{args: %{&quot;id&quot; =&gt; 1}}] = all_enqueued(worker: MyWorker) Note that the final example, using all_enqueued/1, returns a raw list of matching jobs and does not make an assertion by itself. This makes it possible to test using pattern matching at the expense of being more verbose. Example Given a simple module that enqueues a job: defmodule MyApp.Business do def work(args) do args |&gt; Oban.Job.new(worker: MyApp.Worker, queue: :special) |&gt; Oban.insert!() end end The behaviour can be exercised in your test code: defmodule MyApp.BusinessTest do use ExUnit.Case, async: true use Oban.Testing, repo: MyApp.Repo alias MyApp.Business test &quot;jobs are enqueued with provided arguments&quot; do Business.work(%{id: 1, message: &quot;Hello!&quot;}) assert_enqueued worker: MyApp.Worker, args: %{id: 1, message: &quot;Hello!&quot;} end end Adding to Case Templates To include helpers in all of your tests you can add it to your case template: defmodule MyApp.DataCase do use ExUnit.CaseTemplate using do quote do use Oban.Testing, repo: MyApp.Repo import Ecto import Ecto.Changeset import Ecto.Query import MyApp.DataCase alias MyApp.Repo end end end"},{"ref":"Oban.Testing.html#all_enqueued/2","title":"Oban.Testing.all_enqueued/2","type":"function","doc":"Retrieve all currently enqueued jobs matching a set of options. Only jobs matching all of the provided arguments will be returned. Additionally, jobs are returned in descending order where the most recently enqueued job will be listed first. Examples Assert based on only some of a job&#39;s args: assert [%{args: %{&quot;id&quot; =&gt; 1}}] = all_enqueued(worker: MyWorker) Assert that exactly one job was inserted for a queue: assert [%Oban.Job{}] = all_enqueued(queue: :alpha)"},{"ref":"Oban.Testing.html#assert_enqueued/2","title":"Oban.Testing.assert_enqueued/2","type":"function","doc":"Assert that a job with particular options has been enqueued. Only values for the provided arguments will be checked. For example, an assertion made on worker: &quot;MyWorker&quot; will match any jobs for that worker, regardless of the queue or args."},{"ref":"Oban.Testing.html#refute_enqueued/2","title":"Oban.Testing.refute_enqueued/2","type":"function","doc":"Refute that a job with particular options has been enqueued. See assert_enqueued/2 for additional details."},{"ref":"Oban.Worker.html","title":"Oban.Worker","type":"behaviour","doc":"Defines a behavior and macro to guide the creation of worker modules. Worker modules do the work of processing a job. At a minimum they must define a perform/2 function, which will be called with an args map and the job struct. Defining Workers Define a worker to process jobs in the events queue, retrying at most 10 times if a job fails, and ensuring that duplicate jobs aren&#39;t enqueued within a 30 second period: defmodule MyApp.Workers.Business do use Oban.Worker, queue: &quot;events&quot;, max_attempts: 10, unique: [period: 30] @impl Worker def perform(_args, %Oban.Job{attempt: attempt}) when attempt &gt; 3 do IO.inspect(attempt) end def perform(args, _job) do IO.inspect(args) end end The perform/2 function receives an args map and an Oban.Job struct as arguments. This allows workers to change the behavior of perform/2 based on attributes of the Job, e.g. the number of attempts or when it was inserted. A job is considered complete if perform/2 returns a non-error value, and it doesn&#39;t raise an exception or have an unhandled exit. Any of these return values or error events will fail the job: return {:error, error} return :error an unhandled exception an unhandled exit or throw As an example of error tuple handling, this worker may return an error tuple when the value is less than one: defmodule MyApp.Workers.ErrorExample do use Oban.Worker @impl Worker def perform(%{&quot;value&quot; =&gt; value}, _job) do if value &gt; 1 do :ok else {:error, &quot;invalid value given: &quot; &lt;&gt; inspect(value)} end end end Enqueuing Jobs All workers implement a new/2 function that converts an args map into a job changeset suitable for inserting into the database for later execution: %{in_the: &quot;business&quot;, of_doing: &quot;business&quot;} |&gt; MyApp.Workers.Business.new() |&gt; Oban.insert() The worker&#39;s defaults may be overridden by passing options: %{vote_for: &quot;none of the above&quot;} |&gt; MyApp.Workers.Business.new(queue: &quot;special&quot;, max_attempts: 5) |&gt; Oban.insert() Uniqueness options may also be overridden by passing options: %{expensive: &quot;business&quot;} |&gt; MyApp.Workers.Business.new(unique: [period: 120, fields: [:worker]]) |&gt; Oban.insert() Note that unique options aren&#39;t merged, they are overridden entirely. See Oban.Job for all available options. Unique Jobs The unique jobs feature lets you specify constraints to prevent enqueuing duplicate jobs. Uniquness is based on a combination of args, queue, worker, state and insertion time. It is configured at the worker or job level using the following options: :period — The number of seconds until a job is no longer considered duplicate. You should always specify a period. :fields — The fields to compare when evaluating uniqueness. The available fields are :args, :queue and :worker, by default all three are used. :states — The job states that will be checked for duplicates. The available states are :available, :scheduled, :executing, :retryable and :completed. By default all states are checked, which prevents any duplicates, even if the previous job has been completed. For example, configure a worker to be unique across all fields and states for 60 seconds: use Oban.Worker, unique: [period: 60] Configure the worker to be unique only by :worker and :queue: use Oban.Worker, unique: [fields: [:queue, :worker], period: 60] Or, configure a worker to be unique until it has executed: use Oban.Worker, unique: [period: 300, states: [:available, :scheduled, :executing]] Stronger Guarantees Oban&#39;s unique job support is built on a client side read/write cycle. That makes it subject to duplicate writes if two transactions are started simultaneously. If you absolutely must ensure that a duplicate job isn&#39;t inserted then you will have to make use of unique constraints within the database. Oban.insert/2,4 will handle unique constraints safely through upsert support. Performance Note If your application makes heavy use of unique jobs you may want to add indexes on the args and inserted_at columns of the oban_jobs table. The other columns considered for uniqueness are already covered by indexes. Customizing Backoff When jobs fail they may be retried again in the future using a backoff algorithm. By default the backoff is exponential with a fixed padding of 15 seconds. This may be too aggressive for jobs that are resource intensive or need more time between retries. To make backoff scheduling flexible a worker module may define a custom backoff function. This worker defines a backoff function that delays retries using a variant of the historic Resque/Sidekiq algorithm: defmodule MyApp.SidekiqBackoffWorker do use Oban.Worker @impl Worker def backoff(attempt) do :math.pow(attempt, 4) + 15 + :rand.uniform(30) * attempt end @impl Worker def perform(_args, _job) do :do_business end end Here are some alternative backoff strategies to consider: constant — delay by a fixed number of seconds, e.g. 1→15, 2→15, 3→15 linear — delay for the same number of seconds as the current attempt, e.g. 1→1, 2→2, 3→3 squared — delay by attempt number squared, e.g. 1→1, 2→4, 3→9 sidekiq — delay by a base amount plus some jitter, e.g. 1→32, 2→61, 3→135"},{"ref":"Oban.Worker.html#c:backoff/1","title":"Oban.Worker.backoff/1","type":"callback","doc":"Calculate the execution backoff, or the number of seconds to wait before retrying a failed job."},{"ref":"Oban.Worker.html#c:new/2","title":"Oban.Worker.new/2","type":"callback","doc":"Build a job changeset for this worker with optional overrides. See Oban.Job.new/2 for the available options."},{"ref":"Oban.Worker.html#c:perform/2","title":"Oban.Worker.perform/2","type":"callback","doc":"The perform/2 function is called when the job is executed. The value returned from perform/2 is ignored, unless it returns an {:error, reason} tuple. With an error return or when perform has an uncaught exception or throw then the error will be reported and the job will be retried (provided there are attempts remaining)."},{"ref":"readme.html","title":"README","type":"extras","doc":"Robust job processing in Elixir, backed by modern PostgreSQL. Reliable, observable and loaded with enterprise grade features."},{"ref":"readme.html#table-of-contents","title":"README - Table of Contents","type":"extras","doc":"Features Requirements Installation Usage Configuring Queues Defining Workers Enqueuing Jobs Pruning Testing Troubleshooting Contributing"},{"ref":"readme.html#features","title":"README - Features","type":"extras","doc":"Oban&#39;s primary goals are reliability, consistency and observability. It is fundamentally different from other background job processing tools because it retains job data for historic metrics and inspection. You can leave your application running indefinitely without worrying about jobs being lost or orphaned due to crashes. Advantages over in-memory, mnesia, Redis and RabbitMQ based tools: Fewer Dependencies — If you are running a web app there is a very good chance that you&#39;re running on top of a RDBMS. Running your job queue within PostgreSQL minimizes system dependencies and simplifies data backups. Transactional Control — Enqueue a job along with other database changes, ensuring that everything is committed or rolled back atomically. Database Backups — Jobs are stored inside of your primary database, which means they are backed up together with the data that they relate to. Advanced features and advantages over other RDBMS based tools: Isolated Queues — Jobs are stored in a single table but are executed in distinct queues. Each queue runs in isolation, ensuring that a jobs in a single slow queue can&#39;t back up other faster queues. Queue Control — Queues can be paused, resumed and scaled independently at runtime. Resilient Queues — Failing queries won&#39;t crash the entire supervision tree, instead they trip a circuit breaker and will be retried again in the future. Job Killing — Jobs can be killed in the middle of execution regardless of which node they are running on. This stops the job at once and flags it as discarded. Triggered execution — Database triggers ensure that jobs are dispatched as soon as they are inserted into the database. Unique Jobs — Duplicate work can be avoided through unique job controls. Uniqueness can be enforced at the argument, queue and worker level for any period of time. Scheduled Jobs — Jobs can be scheduled at any time in the future, down to the second. Job Safety — When a process crashes or the BEAM is terminated executing jobs aren&#39;t lost—they are quickly recovered by other running nodes or immediately when the node is restarted. Historic Metrics — After a job is processed the row is not deleted. Instead, the job is retained in the database to provide metrics. This allows users to inspect historic jobs and to see aggregate data at the job, queue or argument level. Node Metrics — Every queue records metrics to the database during runtime. These are used to monitor queue health across nodes and may be used for analytics. Queue Draining — Queue shutdown is delayed so that slow jobs can finish executing before shutdown. Telemetry Integration — Job life-cycle events are emitted via Telemetry integration. This enables simple logging, error reporting and health checkups without plug-ins."},{"ref":"readme.html#requirements","title":"README - Requirements","type":"extras","doc":"Oban has been developed and actively tested with Elixir 1.8+, Erlang/OTP 21.1+ and PostgreSQL 11.0+. Running Oban currently requires Elixir 1.8+, Erlang 21+, and PostgreSQL 9.6+."},{"ref":"readme.html#installation","title":"README - Installation","type":"extras","doc":"Oban is published on Hex. Add it to your list of dependencies in mix.exs: def deps do [ {:oban, &quot;~&gt; 0.5&quot;} ] end Then run mix deps.get to install Oban and its dependencies, including Ecto, Jason and Postgrex. After the packages are installed you must create a database migration to add the oban_jobs table to your database: mix ecto.gen.migration add_oban_jobs_table Open the generated migration in your editor and call the up and down functions on Oban.Migrations: defmodule MyApp.Repo.Migrations.AddObanJobsTable do use Ecto.Migration def up do Oban.Migrations.up() end # We specify `version: 1` in `down`, ensuring that we&#39;ll roll all the way back down if # necessary, regardless of which version we&#39;ve migrated `up` to. def down do Oban.Migrations.down(version: 1) end end This will run all of Oban&#39;s versioned migrations for your database. Migrations between versions are idempotent and will never change after a release. As new versions are released you may need to run additional migrations. Now, run the migration to create the table: mix ecto.migrate Next see Usage for how to integrate Oban into your application and start defining jobs! Note About Releases If you are using releases you may see Postgrex errors logged during your initial deploy (or any deploy requiring an Oban migration). The errors are only temporary. After the migration has completed each queue will start producing jobs normally."},{"ref":"readme.html#usage","title":"README - Usage","type":"extras","doc":"Oban isn&#39;t an application and won&#39;t be started automatically. It is started by a supervisor that must be included in your application&#39;s supervision tree. All of your configuration is passed into the Oban supervisor, allowing you to configure Oban like the rest of your application. # config/config.exs config :my_app, Oban, repo: MyApp.Repo, prune: {:maxlen, 100_000}, queues: [default: 10, events: 50, media: 20] # lib/my_app/application.ex defmodule MyApp.Application do @moduledoc false use Application alias MyApp.{Endpoint, Repo} def start(_type, _args) do children = [ Repo, Endpoint, {Oban, Application.get_env(:my_app, Oban)} ] Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor) end end If you are running tests (which you should be) you&#39;ll want to disable pruning and job dispatching altogether when testing: # config/test.exs config :my_app, Oban, queues: false, prune: :disabled Without dispatch and pruning disabled Ecto will raise constant ownership errors and you won&#39;t be able to run tests. Configuring Queues Queues are specified as a keyword list where the key is the name of the queue and the value is the maximum number of concurrent jobs. The following configuration would start four queues with concurrency ranging from 5 to 50: queues: [default: 10, mailers: 20, events: 50, media: 5] There isn&#39;t a limit to the number of queues or how many jobs may execute concurrently. Here are a few caveats and guidelines: Each queue will run as many jobs as possible concurrently, up to the configured limit. Make sure your system has enough resources (i.e. database connections) to handle the concurrent load. Only jobs in the configured queues will execute. Jobs in any other queue will stay in the database untouched. Be careful how many concurrent jobs make expensive system calls (i.e. FFMpeg, ImageMagick). The BEAM ensures that the system stays responsive under load, but those guarantees don&#39;t apply when using ports or shelling out commands. Defining Workers Worker modules do the work of processing a job. At a minimum they must define a perform/2 function, which is called with an args map and the job struct. Define a worker to process jobs in the events queue: defmodule MyApp.Business do use Oban.Worker, queue: &quot;events&quot;, max_attempts: 10 @impl Oban.Worker def perform(%{&quot;id&quot; =&gt; id}, _job) do model = MyApp.Repo.get(MyApp.Business.Man, id) IO.inspect(model) end end The value returned from perform/2 is ignored, unless it returns an {:error, reason} tuple. With an error return or when perform has an uncaught exception or throw then the error will be reported and the job will be retried (provided there are attempts remaining). The Business worker can also be configured to prevent duplicates for a period of time through the :unique option. Here we&#39;ll configure it to be unique for 60 seconds: use Oban.Worker, queue: &quot;events&quot;, max_attempts: 10, unique: [period: 60] Enqueueing Jobs Jobs are simply Ecto structs and are enqueued by inserting them into the database. For convenience and consistency all workers provide a new/2 function that converts an args map into a job changeset suitable for insertion: %{in_the: &quot;business&quot;, of_doing: &quot;business&quot;} |&gt; MyApp.Business.new() |&gt; Oban.insert() The worker&#39;s defaults may be overridden by passing options: %{vote_for: &quot;none of the above&quot;} |&gt; MyApp.Business.new(queue: &quot;special&quot;, max_attempts: 5) |&gt; Oban.insert() Jobs may be scheduled at a specific datetime in the future: %{id: 1} |&gt; MyApp.Business.new(scheduled_at: ~U[2020-12-25 19:00:56.0Z]) |&gt; Oban.insert() Jobs may also be scheduled down to the second any time in the future: %{id: 1} |&gt; MyApp.Business.new(schedule_in: 5) |&gt; Oban.insert() Unique jobs can be configured in the worker, or when the job is built: %{email: &quot;brewster@example.com&quot;} |&gt; MyApp.Mailer.new(unique: [period: 300, fields: [:queue, :worker]) |&gt; Oban.insert() Multiple jobs can be inserted in a single transaction: Ecto.Multi.new() |&gt; Oban.insert(:b_job, MyApp.Business.new(%{id: 1})) |&gt; Oban.insert(:m_job, MyApp.Mailer.new(%{email: &quot;brewser@example.com&quot;})) |&gt; Repo.transaction() Occasionally you may need to insert a job for a worker that exists in another application. In that case you can use Oban.Job.new/2 to build the changeset manually: %{id: 1, user_id: 2} |&gt; Oban.Job.new(queue: :default, worker: OtherApp.Worker) |&gt; Oban.insert() Oban.insert/2,4 is the preferred way of inserting jobs as it provides some of Oban&#39;s advanced features (i.e., unique jobs). However, you can use your application&#39;s Repo.insert/2 function if necessary. Pruning Although Oban keeps all jobs in the database for durability and observability, it&#39;s not a great thing if the table grows indefinitely. Job pruning helps us by deleting old records from the oban_jobs tables. It has 3 modes: Disabled - No jobs are deleted. Example: :disabled Limit-based - Keeps the latest N records. Example: {:maxlen, 100_000} Time-based - Keeps records for the last N seconds. Example for 7 days: {:maxage, 60 * 60 * 24 * 7} Important: Pruning is only applied to jobs that are completed or discarded (has reached the maximum number of retries or has been manually killed). It&#39;ll never delete a new job, a scheduled job or a job that failed and will be retried."},{"ref":"readme.html#testing","title":"README - Testing","type":"extras","doc":"As noted in the Usage section above there are some guidelines for running tests: Disable all job dispatching by setting queues: false or queues: nil in your test.exs config. Keyword configuration is deep merged, so setting queues: [] won&#39;t have any effect. Disable pruning via prune: :disabled. Pruning isn&#39;t necessary in testing mode because jobs created within the sandbox are rolled back at the end of the test. Additionally, the periodic pruning queries will raise DBConnection.OwnershipError when the application boots. Be sure to use the Ecto Sandbox for testing. Oban makes use of database pubsub events to dispatch jobs, but pubsub events never fire within a transaction. Since sandbox tests run within a transaction no events will fire and jobs won&#39;t be dispatched. config :my_app, MyApp.Repo, pool: Ecto.Adapters.SQL.Sandbox Oban provides some helpers to facilitate testing. The helpers handle the boilerplate of making assertions on which jobs are enqueued. To use the assert_enqueued/1 and refute_enqueued/1 helpers in your tests you must include them in your testing module and specify your app&#39;s Ecto repo: use Oban.Testing, repo: MyApp.Repo Now you can assert, refute or list jobs that have been enqueued within your tests: assert_enqueued worker: MyWorker, args: %{id: 1} # or refute_enqueued queue: &quot;special&quot;, args: %{id: 2} # or assert [%{args: %{&quot;id&quot; =&gt; 1}}] = all_enqueued worker: MyWorker See the Oban.Testing module for more details. Integration Testing During integration testing it may be necessary to run jobs because they do work essential for the test to complete, i.e. sending an email, processing media, etc. You can execute all available jobs in a particular queue by calling Oban.drain_queue/1 directly from your tests. For example, to process all pending jobs in the &quot;mailer&quot; queue while testing some business logic: defmodule MyApp.BusinessTest do use MyApp.DataCase, async: true alias MyApp.{Business, Worker} test &quot;we stay in the business of doing business&quot; do :ok = Business.schedule_a_meeting(%{email: &quot;monty@brewster.com&quot;}) assert %{success: 1, failure: 0} == Oban.drain_queue(:mailer) # Now, make an assertion about the email delivery end end See Oban.drain_queue/1 for additional details."},{"ref":"readme.html#troubleshooting","title":"README - Troubleshooting","type":"extras","doc":"Heroku If your app crashes on launch, be sure to confirm you are running the correct version of Elixir and Erlang (view requirements). If using the hashnuke/elixir buildpack, you can update the elixir_buildpack.config file in your application&#39;s root directory to something like: # Elixir version elixir_version=1.9.0 # Erlang version erlang_version=22.0.3 Available Erlang versions are available here."},{"ref":"readme.html#contributing","title":"README - Contributing","type":"extras","doc":"To run the Oban test suite you must have PostgreSQL 10+ running locally with a database named oban_test. Follow these steps to create the database, create the database and run all migrations: # Create the database MIX_ENV=test mix ecto.create -r Oban.Test.Repo # Run the base migration MIX_ENV=test mix ecto.migrate -r Oban.Test.Repo To ensure a commit passes CI you should run mix ci locally, which executes the following commands: Check formatting (mix format --check-formatted) Lint with Credo (mix credo --strict) Run all tests (mix test --raise) Run Dialyzer (mix dialyzer --halt-exit-status)"},{"ref":"CHANGELOG.md.html","title":"CHANGELOG","type":"extras","doc":"Changelog All notable changes to this project will be documented in this file. The format is based on Keep a Changelog and this project adheres to Semantic Versioning."},{"ref":"CHANGELOG.md.html#unreleased","title":"CHANGELOG - Unreleased","type":"extras","doc":""},{"ref":"CHANGELOG.md.html#0-9-0-2019-09-20","title":"CHANGELOG - 0.9.0 — 2019-09-20","type":"extras","doc":"Added [Oban] Add insert_all/2 and insert_all/4, corresponding to Ecto.Repo.insert_all/3 and Ecto.Multi.insert_all/5, respectively. [Oban.Job] Add to_map/1 for converting a changeset into a map suitable for database insertion. This is used by Oban.insert_all/2,4 internally and is exposed for convenience. Changed [Oban.Config] Remove the default queue value of [default: 10], which was overriden by Oban.start_link/1 anyhow. [Oban.Telemetry] Allow the log level to be customized when attaching the default logger. The default level is :info, the same as it was before. Fixed [Oban.Migrations] Prevent invalid up and down targets when attempting to run migrations that have already been ran. This was primarily an issue in CI, where the initial migration was unscoped and would migrate to the current version while a subsequent migration would attempt to migrate to a lower version. [Oban.Job] Prevent a queue comparison with nil by retaining the default queue (default) when building uniqueness checks. [Oban.Job] Set state to scheduled for jobs created with a scheduled_at timestamp. Previously the state was only set when schedule_in was used."},{"ref":"CHANGELOG.md.html#0-8-1-2019-09-11","title":"CHANGELOG - 0.8.1 — 2019-09-11","type":"extras","doc":"Changed [Oban.Notifier] Restore the gossip macro and allow oban_gossip channel for notifications. Fixed [Oban.Migrations] Prevent invalid up and down ranges when repeatedly migrating without specifying a version. This issue was seen when running all of the up migrations on a database from scratch, as there would be multiple oban migrations that simply delegated to up and down."},{"ref":"CHANGELOG.md.html#0-8-0-2019-09-06","title":"CHANGELOG - 0.8.0 — 2019-09-06","type":"extras","doc":"Migration Required (V4) This release requires a migration to V3, with an optional migration to V4. V3 adds a new column to jobs and creates the oban_beats table, while V4 drops a function used by the old advisory locking system. For a smooth zero-downtime upgrade, migrate to V3 first and then V4 in a separate release. The following sample migration will only upgrade to V3: defmodule MyApp.Repo.Migrations.AddObanBeatsTable do use Ecto.Migration def up do Oban.Migrations.up(version: 3) end def down do Oban.Migrations.down(version: 2) end end Added [Oban.Job] Add an attempted_by column used to track the node, queue and producer nonce that attempted to execute a job. [Oban.Beat] Add a new oban_beats table, used by producers to publish &quot;pulse&quot; information including the node, queue, running jobs and other information previously published by gossip notifications. Beat records older than one hour are pruned automatically. Beat pruning respects the :disabled setting, but it ignores length and age configuration. The goal is to prevent bloating the table with useless historic information—each queue generates one beat a second, 3,600 beat records per hour even when the queue is idle. Changed [Oban.Executor] Don&#39;t swallow an ArgumentError when raised by a worker&#39;s backoff function. [Oban.Notifier] Remove gossip notifications entirely, superseded by pulse activity written to oban_beats. [Oban.Query] Remove all use of advisory locks! [Oban.Producer] Periodically attempt to rescue orphans, not just at startup. By default a rescue is attempted once a minute and it checks for any executing jobs belonging to a producer that hasn&#39;t had a pulse for more than a minute. Fixed [Oban.Worker] Validate worker options after the module is compiled. This allows dynamic configuration of compile time settings via module attributes, functions, Application.get_env/3, etc. [Oban.Query] Remove scheduled_at check from job fetching query. This could prevent available jobs from dispatching when the database&#39;s time differed from the system time. [Oban.Migrations] Fix off-by-one error when detecting the version to migrate up from."},{"ref":"CHANGELOG.md.html#0-7-1-2019-08-15","title":"CHANGELOG - 0.7.1 — 2019-08-15","type":"extras","doc":"Fixed [Oban.Query] Release advisory locks in batches rather than individually after a job finishes execution. By tracking unlockable jobs and repeatedly attempting to unlock them for each connection we ensure that eventually all advisory locks are released. The previous unlocking system leaked advisory locks at a rate proportional to the number of connections in the db pool. The more connections, the more locks that wouldn&#39;t release. With a default value of 64 for max_locks_per_transaction the database would raise &quot;ERROR: 53200: out of shared memory&quot; after it hit a threshold (11,937 exactly, in my testing). [Oban.Worker] Allow max_attempts to be 1 or more. This used to be possible and was broken unintentionally by compile time validations."},{"ref":"CHANGELOG.md.html#0-7-0-2019-08-08","title":"CHANGELOG - 0.7.0 — 2019-08-08","type":"extras","doc":"Added [Oban] Added insert/2, insert!/2 and insert/4 as a convenient and more powerful way to insert jobs. Features such as unique jobs and the upcoming prefix support only work with insert. [Oban] Add prefix support. This allows entirely isolated job queues within the same database. [Oban.Worker] Compile time validation of all passed options. Catch typos and other invalid options when a worker is compiled rather than when a job is inserted for the first time. [Oban.Worker] Unique job support through the unique option. Set a unique period, and optionally fields and states, to enforce uniqueness within a window of time. For example, to make a job unique by args, queue and worker for 2 minutes: use Oban.Worker, unique: [period: 120, fields: [:args, :queue, :worker]] Note, unique support relies on the use of Oban.insert/2,4. Changed [Oban.Worker] Remove the perform/1 callback in favor of perform/2. The new perform/2 function receives the job&#39;s args, followed by the complete job struct. This new function signature makes it clear that the args are always available, and the job struct is also there when it is needed. A default perform/2 function is not generated automatically by the use macro and must be defined manually. This is a breaking change and all worker modules will need to be updated. Thankfully, due to the behaviour change, warnings will be emitted when you compile after the upgrade. If your perform functions weren&#39;t matching on the Oban.Job struct then you can migrate your workers by adding a second _job argument: def perform(%{&quot;some&quot; =&gt; args}, _job) If you were making use of Oban.Job metadata in perform/1 then you can move the job matching to the second argument: def perform(_args, %{attempt: attempt}) See the issue that suggested this change for more details and discussion. [Oban.Producer] Use send_after/3 instead of :timer.send_interval/2 to maintain scheduled dispatch. This mechanism is more accurate under system load and it prevents :poll messages from backing up for each producer. [Oban.Migration] Accept a keyword list with :prefix and :version as options rather than a single version string. When a prefix is supplied the migration will create all tables, indexes, functions and triggers within that namespace. For example, to create the jobs table within a &quot;private&quot; prefix: Oban.Migrate.up(prefix: &quot;private&quot;)"},{"ref":"CHANGELOG.md.html#0-6-0-2019-07-26","title":"CHANGELOG - 0.6.0 — 2019-07-26","type":"extras","doc":"Added [Oban.Query] Added :verbose option to control general query logging. This allows debug query activity within Oban to be silenced during testing and development. [Oban.Testing] Added all_enqueued/1 helper for testing. The helper returns a list of jobs matching the provided criteria. This makes it possible to test using pattern matching, which is more flexible than a literal match within the database. Changed [Oban.Config] All passed options are validated. Any unknown options will raise an ArgumentError and halt startup. This prevents misconfiguration through typos and passing unsupported options. [Oban.Worker] The perform/1 function now receives an Oban.Job struct as the sole argument, calling perform/1 again with only the args map if no clause catches the struct. This allows workers to use any attribute of the job to customize behaviour, e.g. the number of attempts or when a job was inserted into the database. The implementation is entirely backward compatible, provided workers are defined with the use macro. Workers that implement the Oban.Worker behaviour manually will need to change the signature of perform/1 to accept a job struct. [Oban] Child process names are generated from the top level supervisor&#39;s name, i.e. setting the name to &quot;MyOban&quot; on start_link/1 will set the notifier&#39;s name to MyOban.Notifier. This improves isolation and allows multiple supervisors to be ran on the same node. Fixed [Oban.Producer] Remove duplicate polling timers. As part of a botched merge conflict resolution two timers were started for each producer."},{"ref":"CHANGELOG.md.html#0-5-0-2019-06-27","title":"CHANGELOG - 0.5.0 — 2019-06-27","type":"extras","doc":"Added [Oban.Pruner] Added :prune_limit option to constrain the number of rows deleted on each prune iteration. This prevents locking the database when there are a large number of jobs to delete. Changed [Oban.Worker] Treat {:error, reason} tuples returned from perform/1 as a failure. The :kind value reported in telemetry events is now differentiated, where a rescued exception has the kind :exception, and an error tuple has the kind :error. Fixed [Oban.Testing] Only check available and scheduled jobs with the assert|refute_enqueued testing helpers. [Oban.Queue.Watchman] Catch process exits when attempting graceful shutdown. Any exits triggered within terminate/2 are caught and ignored. This safely handles situations where the producer exits before the watchman does. [Oban.Queue.Producer] Use circuit breaker protection for gossip events and call queries directly from the producer process. This prevents pushing all queries through the Notifier, and ensures more granular control over gossip errors."},{"ref":"CHANGELOG.md.html#0-4-0-2019-06-10","title":"CHANGELOG - 0.4.0 — 2019-06-10","type":"extras","doc":"Added [Oban] Add Oban.drain_queue/1 to help with integration testing. Draining a queue synchronously executes all available jobs in the queue from within the calling process. This avoids any sandbox based database connection issues and prevents race conditions from asynchronous processing. [Oban.Worker] Add backoff/1 callback and switch to exponential backoff with a base value as the default. This allows custom backoff timing for individual workers. [Oban.Telemetry] Added a new module to wrap a default handler for structured JSON logging. The log handler is attached by calling Oban.Telemetry.attach_default_logger/0 somewhere in your application code. [Oban.Queue.Producer] Guard against Postgrex errors in all producer queries using a circuit breaker. Failing queries will no longer crash the producer. Instead, the failure will be logged as an error and it will trip the producer&#39;s circuit breaker. All subsequent queries will be skipped until the breaker is enabled again approximately a minute later. This feature simplifies the deployment process by allowing the application to boot and stay up while Oban migrations are made. After migrations have finished each queue producer will resume making queries. Changed [Oban] Telemetry events now report timing as %{duration: duration} instead of %{timing: timing}. This aligns with the telemetry standard of using duration for the time to execute something. [Oban] Telemetry events are now broken into success and failure at the event level, rather than being labeled in the metadata. The full event names are now [:oban, :success] and [:oban, :failure]. [Oban.Job] Rename scheduled_in to schedule_in for readability and consistency. Both the Oban docs and README showed schedule_in, which reads more clearly than scheduled_in. [Oban.Pruner] Pruning no longer happens immediately on startup and may be configured through the :prune_interval option. The default prune interval is still one minute. Fixed [Oban.Migrations] Make partial migrations more resilient by guarding against missing versions and using idempotent statements."},{"ref":"CHANGELOG.md.html#0-3-0-2019-05-29","title":"CHANGELOG - 0.3.0 - 2019-05-29","type":"extras","doc":"Migration Required (V2) Added [Oban] Allow setting queues: false or queues: nil to disable queue dispatching altogether. This makes it possible to override the default configuration within each environment, i.e. when testing. The docs have been updated to promote this mechanism, as well as noting that pruning must be disabled for testing. [Oban.Testing] The new testing module provides a set of helpers to make asserting and refuting enqueued jobs within tests much easier. Changed [Oban.Migrations] Explicitly set id as a bigserial to avoid mistakenly generating a uuid primary key. [Oban.Migrations] Use versioned migrations that are immutable. As database changes are required a new migration module is defined, but the interface of Oban.Migrations.up/0 and Oban.Migrations.down/0 will be maintained. From here on all releases with database changes will indicate that a new migration is necessary in this CHANGELOG. [Oban.Query] Replace use of (bigint) with (int, int) for advisory locks. The first int acts as a namespace and is derived from the unique oid value for the oban_jobs table. The oid is unique within a database and even changes on repeat table definitions. This change aims to prevent lock collision with application level advisory lock usage and other libraries. Now there is a 1 in 2,147,483,647 chance of colliding with other locks. [Oban.Job] Automatically remove leading &quot;Elixir.&quot; namespace from stringified worker name. The prefix complicates full text searching and reduces the score for trigram matches."},{"ref":"CHANGELOG.md.html#0-2-0-2019-05-15","title":"CHANGELOG - 0.2.0 - 2019-05-15","type":"extras","doc":"Added [Oban] Add pause_queue/2, resume_queue/2 and scale_queue/3 for dynamically controlling queues. [Oban] Add kill_job/2 for terminating running jobs and marking them as discarded. [Oban] Add config/0 for retreiving the supervisors config. This is primarily useful for integrating oban into external applications. [Oban.Queue.Producer] Use database triggers to immediately dispatch when a job is inserted into the oban_jobs table. [Oban.Job] Execution errors are stored as a jsonb array for each job. Each error is stored, not just the most recent one. Error entries contains these keys: at The utc timestamp when the error occurred at attempt The attempt number when the error ocurred error A formatted error message and stacktrace, passed through Exception.blame/3 [Oban.Config] Validate all options based on type and allowed values. Any invalid option will raise, preventing supervisor boot. [Oban.Notifier] Broadcast runtime gossip through pubsub, allowing any external system to get stats at the node and queue level. Changed [Oban.Queue.Supervisor] Set the min_demand to 1 for all consumers. This ensures that each queue will run the configured number of concurrent jobs. By default the min_demand is half of max_demand, which means a few slow jobs can prevent the queue from running the expected number of concurrent jobs. [Oban.Job] Change psuedo-states based on job properties into fixed states, this applies to scheduled and retryable. [Oban.Job] The &quot;Elixir&quot; prefix is stripped from worker names before storing jobs in the database. Module lookup performs the same way, but it cleans up displaying the worker name as a string. [Oban.Job] Accept all job fields as changeset parameters. While not encouraged for regular use, this is essential for testing various states."},{"ref":"CHANGELOG.md.html#0-1-0-2019-03-10","title":"CHANGELOG - 0.1.0 - 2019-03-10","type":"extras","doc":"[Oban] Initial release with base functionality."}]